#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <fcntl.h>
#include <errno.h>
#include <unistd.h>
#include <arpa/inet.h>
#include <ctype.h>
#include "ataraw.h"
#ifdef HAVE_ERR_H
#include <err.h>
#endif

#ifndef HAVE_ERR
#include <stdarg.h>
void err(int eval,const char *fmt,...)
{
  va_list ap;
  va_start(ap,fmt);
  vfprintf(stderr,fmt,ap);
  va_end(ap);
  fprintf(stderr,": %s\n",strerror(errno));
  exit(eval);
}
#endif
//some print commands
void print_identify(u_char buff[512] );
void print_dco_identify( dco_struct* dco );

const char *progname="main"; 

void usage()
{
    fprintf(stderr,"usage: %s <device> <command> [args]\n",progname);
    puts("supported commands:");
    puts("  test          - runs the selftest");   
    puts("  id, identify  - send IDENTIFY and other status commands to drive and print results");
    puts("  commands      - reports which ATA commands are supported by this program");
    puts("  read nnn      - read and print a 28-bit sector nnn with PIO");
    puts("  read48 nnn    - read and print a 48-bit sector number with PIO");
    puts("  readdma nnn   - read and print a 28-bit sector nnn with DMA");
    puts("  read48dma nnn - read and print a 48-bit sector nnn with DMA");
    puts("  readbuffer    - read the devices sector buffer");
    puts("  setmax nnn    - CAREFUL: sets the maximum addressable sector on the device(28 bit).");
    puts("  setmaxext nnn - CAREFUL: sets the maximum addressable sector on the device(48 bit).");
    puts("  write nnn     - writes a sector of 1's to the device.  This is the 28 bit version.");
    puts("  writeext nnn  - writes a sector of 1's to the device.  This is the 48 bit version.");
    puts("  dcoid         - Peforms a DCO identify command.");
    puts("  dcoset nnn    - CAREFUL: sets the maximum adressable sector via dco (48 bit).");
    puts("  dcorestore    - Resotores the DCO to default settings.");
    puts("  smart         - Enable SMART operations and print the log.\n");
   exit(1);
}

void process_command(int fd,int argc,char **argv)
{
    const char *command = argv[0];
    unsigned char* buffer = calloc( 512,1 );

    argv++;
    argc--;

    if(!strcmp(command,"id") || !strcmp(command,"identify") || !strcmp(command,"test")){
      u_long address = 0;
      uint64_t address48 = 0;

#define words(x) ((buffer[x*2])+(buffer[(x*2)+1]<<8))
      if(ata_identify_device( fd, buffer )==ATA_RAW_OK ){
          
          printf("**Hex Dump of Identify**\n");
          ata_print_sector( buffer );
	  print_identify( buffer );
	} else err(1,"identify_device_ata failed");
      
      if(ata_read_native_max_address(fd,&address)==ATA_RAW_OK){
	printf("read native max address: %lu\n",address);
      } else err(1,"read native max address failed");

      if(ata_read_native_max_ext(fd,&address48)==ATA_RAW_OK){
	printf("ext read native max: %qu  %s\n",address48,
	       address48 >= 0x1000000 ? "(bigger than 28 bits)" : "(smaller than 28 bits)");
      } else err(1,"ext read native max address failed");

      if(strcmp(command,"test")!=0) return;

      /* If we are testing, continue on */
      /* Try the 28-bit read commands*/
      for(address=1;address<0x0FFFFFFF;address++){
	  u_char buf1[512],buf2[512];
	  int result2;
	  printf("Attempting to read address %lx using read()...\n",address);
	  if(lseek(fd,address*512,0)<0){
	      printf("Lseek returned an error; you must have hit the end of the disk.\n");
	      break;
	  }
	  if(read(fd,buf1,512)!=512){
	      printf("read did not return 512 bytes; you must have hit the end of the disk.\n");
	  }
	  printf("Attempting to read address %lx using raw ATA...\n",address);
	  result2 = ata_read_sector(fd,address,1,buf2);
	  if(result2!=ATA_RAW_OK){
	      err(1,"ata_read_sector() returned %d; it should have returned %d\n",result2,ATA_RAW_OK);
	  }
	  if(memcmp(buf1,buf2,512)!=0){
	      err(1,"got different results from the two read commands. Something is wrong.");
	  }
      }

      return;
    }
    if(!strcmp(command,"commands")){
	int i;
	u_char sup[256];
	get_supported_commands(sup);
	for(i=0;i<sizeof(sup) && sup[i]!=0xff;i++){
	    printf("command 0x%02x\n",sup[i]);
	}
	return;
    }
    if(!strcmp( command, "setmax" ) ){
        if( argc<1) usage();
	int max_address = atoi(argv[0]);
	int result = ata_set_max_address( fd, max_address, 0 );
	if( result == ATA_RAW_OK ){
  	    printf("New Maximum sector is %d\n", max_address );
	    return;
	}
	err(1,"Failed to set max address, run id to see if HPA is supported\n");
	return;
    } 
    if(!strcmp( command, "setmaxext" ) ){
        if( argc<1) usage();
	uint64_t max_address = atoll(argv[0]);
	int result = ata_set_max_address_ext( fd, max_address, 0 );
	if( result == ATA_RAW_OK ){
  	    printf("New Maximum sector is %llu\n", max_address );
	    return;
	}
	err(1,"Failed to set max address, run id to see if HPA is supported\n");
	return;
    }
    if(!strcmp(command,"read")){
	u_char buffer[512];
	if(argc<1) usage();
	int sector_number = atoi(argv[0]);
	int result = ata_read_sector( fd, sector_number, 1, buffer ) ;


	if(result == ATA_RAW_OK ){
	    printf("Sector %d:\n",sector_number);
	    ata_print_sector( buffer );
	    return;
	}
	err(1,"FAILED, RETURN CODE: %d\n", result );
	return;
    }
    if(!strcmp(command, "write")){
        u_char buffer[512];
	if( argc<1) usage();
	int sector_number = atoi(argv[0]);
	memset( buffer, 1, sizeof(buffer));
	int result = ata_write_sector( fd,sector_number, 1,buffer);
	if( result == ATA_RAW_OK){
	    printf("Wrote 512 1's to sector %d\n", sector_number );
	    return;
	}
	err(1,"FAILED, RETURN CODE: %d\n", result );
	return;
    }
    if(!strcmp(command, "writeext")){
        u_char buffer[512];
	if( argc<1) usage();
	int sector_number = atoi(argv[0]);
	memset( buffer, 1, sizeof(buffer));
	int result = ata_write_sector_ext( fd, sector_number, 1, buffer );
	if( result == ATA_RAW_OK ){
	    printf("Wrote 512 1's to sector %d\n", sector_number );
	    return;
	}
	err(1,"FAILED, RETURN CODE: %d\n", result );
	return;
    }
    if( !strcmp( command, "readbuffer")){
        u_char buffer[SECTOR_SIZE];
	memset( &buffer, 0, 8 );
	int result = ata_read_buffer( fd, buffer );
	if( result == ATA_RAW_OK ){
	    ata_print_sector( buffer );
	    return;
	}
	err(1,"FAILED, RETURN CODE: %d\n", result );
    }
    if(!strcmp( command, "dcoid" ) ){
    	dco_struct dco;
	int result = ata_device_configuration_identify( fd, (u_char*)&dco );
	if( result == ATA_RAW_OK ){
	    print_dco_identify(&dco);
	    return;
	}
	err(1,"FAILED, RETURN CODE: %d\n", result );
    } 
    if( !strcmp( command, "readdma")){
        u_char buffer[SECTOR_SIZE];
	if(argc<1) usage();
	int sector_number = atoi(argv[0]);
	memset( buffer, 0, sizeof(buffer));
	int result = ata_read_dma( fd, sector_number, buffer );
	if( result == ATA_RAW_OK ){
	    ata_print_sector( buffer );
	    return;
	}
	err(1,"FAILED, RETURN CODE: %d\n", result );
    }
    if(!strcmp(command,"read48")){
        u_char buffer[SECTOR_SIZE];
	if(argc<1) usage();
	uint64_t sector_number = atoll(argv[0]);
	memset( &buffer, 0, sizeof(buffer) );
	printf("WOOT\n");
	int result = ata_read_sector_ext(fd, sector_number, buffer);
	if( result == ATA_RAW_OK ){
            ata_print_sector( buffer );
	    return;
	}
	err(1,"FAILED, RETURN CODE: %d\n", result );
    }
    if(!strcmp(command,"read48dma")){
        u_char buffer[SECTOR_SIZE];
	if(argc<1) usage();
	uint64_t sector_number = atoll(argv[0]);
	memset( &buffer, 0, sizeof(buffer) );
	printf("WOOT\n");
	int result = ata_read_dma_ext(fd, sector_number, buffer);
	if( result == ATA_RAW_OK ){
            ata_print_sector( buffer );
	    return;
	}
	err(1,"FAILED, RETURN CODE: %d\n", result );
    }
    if(!strcmp(command,"dcoset")){
	if(argc<1) usage();
	uint64_t sector_number = atoll(argv[0]);
	int result = ata_dco_set_max_lba(fd, sector_number );
	if( result == ATA_RAW_OK ){
	    printf("Max address is now %llu\n", sector_number );
	    return;
	}
	err(1,"FAILED, RETURN CODE: %d\n", result );
    }
    if(!strcmp(command,"dcorestore")){
       int result = ata_dco_restore( fd );
       if( result == ATA_RAW_OK ){
           printf("DCO Restored\n");
	   return;
       }
       err(1,"FAILED, RETURN CODE: %d\n", result );
    }
    if(!strcmp(command,"smart")){
	if(ata_smart_enable_operations(fd)==ATA_RAW_OK){
	    printf("SMART enabled\n");
	}
	else {
	    printf("Could not enable SMART\n");
	}
	return;
    }
    err(1,"unknown command: '%s'\n",command);
}

int main(int argc,char **argv)
{
    int opt;

    progname = argv[0];
    while((opt=getopt(argc,argv,"d"))!=-1){
	switch(opt){
	case 'd':
	    ataraw_debug = 1;
	    break;
	default:
	    usage();
	}
    }

    argc -= optind;
    argv += optind;

    if(argc<2) usage();

    const char *device = argv[0];

    argv++;
    argc--;

    int fd = open(device,O_RDWR);
    if(fd<0){
	perror(device);
	exit(1);
    }

    /* Now process the command */
    process_command(fd,argc,argv);
    close(fd);
    return 0;
}


void print_identify(u_char buffer[512] ){

    int i;
    // You are invited to finish this...
    // http://www.t10.org/t13/project/d1153r18-ATA-ATAPI-4.pdf
    // http://www.t13.org/Documents/UploadedDocuments/project/d1410r3b-ATA-ATAPI-6.pdf
	  
    printf("\n**Identification information**\n\n");
    printf("Logical Cylinders: %d\n",words(1));
    printf("Logical heads: %d\n",words(3));
    printf("Logical sectors per track: %d\n",words(6));
    printf("Serial number: "); for(i=10;i<=19;i++) printf("%c%c",buffer[(i*2)+1],buffer[i*2]); putchar('\n');
    printf("Firmware Version: "); for(i=23;i<=26;i++) printf("%c%c",buffer[(i*2)+1],buffer[i*2]); putchar('\n');
    printf("Model Number: "); for(i=27;i<=46;i++) printf("%c%c",buffer[(i*2)+1],buffer[i*2]); putchar('\n');
    printf("Max LBA for device (28 bit): %u\n",  words(60)+( words(61)<<16) );
    printf("Max LBA for device (48 bit): %llu\n", words(100)+( words(101)<<16 ) + ((uint64_t) words(102) << 32)+ ((uint64_t)words(103) << 48 ) );
    //device capabilities..words 49-50
    printf("\n**Device Capabilities**\n\n");
    printf("Max number of sectors on READ/WRITE MULTIPLE commands: %d\n",words(47) & 0x7f);
    printf("Current setting for number fo sectors on R/W MULTIPLE: %d\n",words(59) & 0xff);
    printf("Multiple sector setting:  %s\n", (0x010000 & words(59) )?"VALID":"NOT VALID");
    printf("Stanby Timers are:  %s\n", (0x2000 & words(49) )?"SPECIFIED BY THE STANDARD":"HANDLED BY THE DEVICE" );
    printf("IORDY:  %s\n", (0x0800&words(49))?"SUPPORTED":"MAY BE SUPPORTED" );
    printf("IORDY:  %s\n", (0x0400&words(49))?"MAY BE DISABLED":"NOT DISABLED");
    printf("LBA:  %s\n", (0x0200&words(49))?"SUPPORTED":"NOT SUPPORTED");
    printf("DMA:  %s\n", (0x0100&words(49))?"SUPPORTED":"NOT SUPPORTED");
    printf("Multiword DMA Mode 2:  %s\n", (0x040000 & words(63) )?"SELECTED":"NOT SELECTED");
    printf("Multiword DMA Mode 1:  %s\n", (0x020000 & words(63) )?"SELECTED":"NOT SELECTED");
    printf("Multiword DMA Mode 0:  %s\n", (0x010000 & words(63) )?"SELECTED":"NOT SELECTED");
    printf("Multiword DMA Mode 2 and below:    %s\n", (0x04 & words(63) )?"SUPPORTED":"NOT SUPPORTED");  
    printf("Multiword DMA Mode 1 and below:    %s\n", (0x02 & words(63) )?"SUPPORTED":"NOT SUPPORTED");
    printf("Multiword DMA Mode 0:  %s\n", (0x01 & words(63) )?"SUPPORTED":"NOT SUPPORTED");  
    if( words(53) & 0x04 ){
        printf("Ultra DMA Mode 5:      %s\n", ( (words(88)>>8) & 0x20)?"SELECTED":"NOT SELECTED");  
        printf("Ultra DMA Mode 4:      %s\n", ( (words(88)>>8) & 0x10)?"SELECTED":"NOT SELECTED");  
        printf("Ultra DMA Mode 3:      %s\n", ( (words(88)>>8) & 0x08)?"SELECTED":"NOT SELECTED");  
        printf("Ultra DMA Mode 2:      %s\n", ( (words(88)>>8) & 0x04)?"SELECTED":"NOT SELECTED");  
        printf("Ultra DMA Mode 1:      %s\n", ( (words(88)>>8) & 0x02)?"SELECTED":"NOT SELECTED");  
        printf("Ultra DMA Mode 0:      %s\n", ( (words(88)>>8) & 0x01)?"SELECTED":"NOT SELECTED"); 
        printf("Ultra DMA Mode 5 and below:        %s\n",( words(88) & 0x20)?"SUPPORTED":"NOT SUPPORTED");   
        printf("Ultra DMA Mode 4 and below:        %s\n",( words(88) & 0x10)?"SUPPORTED":"NOT SUPPORTED");   
        printf("Ultra DMA Mode 3 and below:        %s\n",( words(88) & 0x08)?"SUPPORTED":"NOT SUPPORTED");   
        printf("Ultra DMA Mode 2 and below:        %s\n",( words(88) & 0x04)?"SUPPORTED":"NOT SUPPORTED");   
        printf("Ultra DMA Mode 1 and below:        %s\n",( words(88) & 0x02)?"SUPPORTED":"NOT SUPPORTED");   
        printf("Ultra DMA Mode 0:      %s\n",( words(88) & 0x01)?"SUPPORTED":"NOT SUPPORTED");   
    }
    printf("Queue Depth:  %d\n", 0x1f & words(75));

    printf("\n**Supported Command Versions**\n\n");
    if(words(80)==0 || words(80)==0xFFFF) printf("device does not report version\n");
    if(words(80) & 0x01) printf("Supports ATA-3\n");
    if(words(80) & 0x10) printf("Supports ATA-4\n"); 
    if(words(80) & 0x20) printf("Supports ATA-5\n");
    if(words(80) & 0x40) printf("Supports ATA-6\n"); 
    if(words(80) & 0x80) printf("Supports ATA-7\n"); 
    if(words(80) & 0x010000) printf("Supports ATA-8\n"); 
    if(words(81)==0 || words(81)==0xFFFF) printf("device does not report minor version\n");
    else printf("minor version number: %d\n",words(81));
	  
    printf("\n**Supported Commands**\n\n");
    printf("NOP Command:          %s\n", ( (words(82)>>8)& 0x40 )?"SUPPORTED":"NOT SUPPORTED" ); 
    printf("READ BUFFER  Command: %s\n", ( (words(82)>>8)& 0x20 )?"SUPPORTED":"NOT SUPPORTED" );  
    printf("WRITE BUFFER Command: %s\n", ( (words(82)>>8)& 0x10 )?"SUPPORTED":"NOT SUPPORTED" );
    printf("HPA Feature Set:      %s\n", ( (words(82)>>8 )& 0x04 )?"SUPPORTED":"NOT SUPPORTED" ); 
    printf("Device Reset Command: %s\n", ( (words(82)>>8 )& 0x02 )?"SUPPORTED":"NOT SUPPORTED" ); 
    printf("Service Interrupt:    %s\n", ( (words(82)>>8 )& 0x01 )?"SUPPORTED":"NOT SUPPORTED" ); 
    printf("Release Interrupt:    %s\n", ( words(82)& 0x80 )?"SUPPORTED":"NOT SUPPORTED" ); 
    printf("Look-Ahead:      %s\n", ( words(82)& 0x40 )?"SUPPORTED":"NOT SUPPORTED" ); 
    printf("Write Cache:     %s\n", ( words(82)& 0x20 )?"SUPPORTED":"NOT SUPPORTED" ); 
    printf("Packet Command Feature Set:            %s\n", ( words(82)& 0x10 )?"NOT SUPPORTED":"SUPPORTED" ); 
    printf("Mandatory Power Mangement Feature Set: %s\n", ( words(82)& 0x08 )?"SUPPORTED":"NOT SUPPORTED" ); 
    printf("Removeable Media Feature Set:          %s\n", ( words(82)& 0x04 )?"SUPPORTED":"NOT SUPPORTED" ); 
    printf("Security Mode Feature Set:             %s\n", ( words(82)& 0x02 )?"SUPPORTED":"NOT SUPPORTED" ); 
    printf("SMART Feature Set:    %s\n", ( words(82)& 0x01 )?"SUPPORTED":"NOT SUPPORTED" );
  
    printf("FLUSH CACHE EXT Command:        %s\n", ( (words(83)>>8)& 0x20 )?"SUPPORTED":"NOT SUPPORTED" );  
    printf("Mandoatory FLUSH CACHE Command: %s\n", ( (words(83)>>8)& 0x10 )?"SUPPORTED":"NOT SUPPORTED" );
    printf("Device Configuration Overlay Feature Set: %s\n", ( (words(83)>>8 )& 0x08 )?"SUPPORTED":"NOT SUPPORTED" ); 
    printf("48-bit Address Feature Set:     %s\n", ( (words(83)>>8 )& 0x04 )?"SUPPORTED":"NOT SUPPORTED" ); 
    printf("Automatic Acoustic Management Feature Set: %s\n", ( (words(83)>>8 )& 0x02 )?"SUPPORTED":"NOT SUPPORTED" ); 
    printf("Set Max Security Extension:     %s\n", ( (words(83)>>8)& 0x01 )?"SUPPORTED":"NOT SUPPORTED" ); 
    printf("Set Features Subcommand required to spin-up at powerup: %s\n", ( words(83)& 0x40 )?"SUPPORTED":"NOT SUPPORTED" ); 
    printf("Power Up Standby Feature Set:   %s\n", ( words(83)& 0x20 )?"SUPPORTED":"NOT SUPPORTED" ); 
    printf("Removeable Media Status Notification Feature Set: %s\n", ( words(83)& 0x10 )?"SUPPORTED":"NOT SUPPORTED" ); 
    printf("Advanced Power Management Feature Set:            %s\n", ( words(83)& 0x08 )?"SUPPORTED":"NOT SUPPORTED" ); 
    printf("CFA Feature Set:               %s\n", ( words(83)& 0x04 )?"SUPPORTED":"NOT SUPPORTED" ); 
    printf("Read/Write DMA Queued:         %s\n", ( words(83)& 0x02 )?"SUPPORTED":"NOT SUPPORTED" ); 
    printf("Download Microcode Command:    %s\n", ( words(83)& 0x01 )?"SUPPORTED":"NOT SUPPORTED" );

    printf("General Purpose Logging Feature Set:   %s\n", ( words(84)& 0x20 )?"SUPPORTED":"NOT SUPPORTED" ); 
    printf("Media Card Pass Thorugh Feature Set:   %s\n", ( words(84)& 0x08 )?"SUPPORTED":"NOT SUPPORTED" ); 
    printf("Media Serial Number:  %s\n", ( words(84)& 0x04 )?"SUPPORTED":"NOT SUPPORTED" ); 
    printf("SMART Self Test:       %s\n", ( words(84)& 0x02 )?"SUPPORTED":"NOT SUPPORTED" ); 
    printf("SMART Error Logging:   %s\n", ( words(84)& 0x01 )?"SUPPORTED":"NOT SUPPORTED" );

    printf("\n**Enabled Commands**\n\n");
    printf("NOP Command:           %s\n", ( (words(85)>>8)& 0x40 )?"ENABLED":"NOT ENABLED" ); 
    printf("Read Buffer Command:   %s\n", ( (words(85)>>8)& 0x20 )?"ENABLED":"NOT ENABLED" ); 
    printf("Write Buffer Command:  %s\n", ( (words(85)>>8)& 0x10 )?"ENABLED":"NOT ENABLED" ); 
    printf("HPA Feature Set:       %s\n", ( (words(85)>>8)& 0x04 )?"ENABLED":"NOT ENABLED" ); 
    printf("Device Reset Command:  %s\n", ( (words(85)>>8)& 0x02 )?"ENABLED":"NOT ENABLED" ); 
    printf("Service Interrupt:     %s\n", ( (words(85)>>8)& 0x01 )?"ENABLED":"NOT ENABLED" ); 
    printf("Release Interrupt:     %s\n", ( words(85)& 0x80 )?"ENABLED":"NOT ENABLED" ); 
    printf("Look-Ahead:            %s\n", ( words(85)& 0x40 )?"ENABLED":"NOT ENABLED" ); 
    printf("Write Cache:           %s\n", ( words(85)& 0x20 )?"ENABLED":"NOT ENABLED" ); 
    printf("Packet Feature Set:    %s\n", ( words(85)& 0x10 )?"NOT ENABLED":"ENABLED" ); 
    printf("Power Management Feature Set:  %s\n", ( words(85)& 0x08 )?"ENABLED":"NOT ENABLED" ); 
    printf("Removeable Feature Set:        %s\n", ( words(85)& 0x04 )?"ENABLED":"NOT ENABLED" ); 
    printf("Security Mode Feature Set:     %s\n", ( words(85)& 0x02 )?"ENABLED":"NOT ENABLED" ); 
    printf("SMART Feature Set:     %s\n", ( words(85)& 0x01 )?"ENABLED":"NOT ENABLED" ); 
   
    printf("FLUSH CACHE EXT Command:        %s\n", ( (words(86)>>8)& 0x20 )?"ENABLED":"NOT ENABLED" );  
    printf("Mandoatory FLUSH CACHE Command: %s\n", ( (words(86)>>8)& 0x10 )?"ENABLED":"NOT ENABLED" );
    printf("Device Configuration Overlay Feature Set: %s\n", ( (words(86)>>8 )& 0x08 )?"ENABLED":"NOT ENABLED" ); 
    printf("48-bit Address Feature Set:     %s\n", ( (words(86)>>8 )& 0x04 )?"ENABLED":"NOT ENABLED" ); 
    printf("Automatic Acoustic Management Feature Set: %s\n", ( (words(86)>>8 )& 0x02 )?"ENABLED":"NOT ENABLED" ); 
    printf("Set Max Security Extension:     %s\n", ( (words(86)>>8)& 0x01 )?"ENABLED":"NOT ENABLED" ); 
    printf("Set Features Subcommand required to spin-up at powerup: %s\n", ( words(86)& 0x40 )?"ENABLED":"NOT ENABLED" ); 
    printf("Power Up Standby Feature Set:   %s\n", ( words(86)& 0x20 )?"ENABLED":"NOT ENABLED" ); 
    printf("Removeable Media Status Notification Feature Set: %s\n", ( words(86)& 0x10 )?"ENABLED":"NOT ENABLED" ); 
    printf("Advanced Power Management Feature Set:            %s\n", ( words(86)& 0x08 )?"ENABLED":"NOT ENABLED" ); 
    printf("CFA Feature Set:               %s\n", ( words(86)& 0x04 )?"ENABLED":"NOT ENABLED" ); 
    printf("Read/Write DMA Queued:         %s\n", ( words(86)& 0x02 )?"ENABLED":"NOT ENABLED" ); 
    printf("Download Microcode Command:    %s\n", ( words(86)& 0x01 )?"ENABLED":"NOT ENABLED" );
   
    printf("General Purpose Logging Feature Set:   %s\n", ( words(87)& 0x20 )?"ENABLED":"NOT ENABLED" ); 
    printf("Media Card Pass Thorugh Feature Set:   %s\n", ( words(87)& 0x08 )?"ENABLED":"NOT ENABLED" ); 
    printf("Media Seerial Number:  %s\n", ( words(87)& 0x04 )?"ENABLED":"NOT ENABLED" ); 
    printf("SMART Self Test:       %s\n", ( words(87)& 0x02 )?"ENABLED":"NOT ENABLED" ); 
    printf("SMART Error Logging:   %s\n", ( words(87)& 0x01 )?"ENABLED":"NOT ENABLED" );

    if( (words(83)>>8) & 0x02 ){ 
        printf("\n**Acoustic Management Feature Set**\n\n");
	printf("Vendor Recommended Acoustic Managment Value: 0x%x\n", (words(94) >> 8) & 0xff );
        printf("Current Automatic Acoustic Management Value: 0x%x\n", words(94) & 0xff );
    }
    if( words(83)& 0x10){
        printf("\n**Removeable Media Status Notification Feature Set Support**\n\n");
	printf("Removeable Media Status Notification Feature Set:  %s\n", (words(127) & 0x01)?"SUPPORTED":"NOT SUPPORTED");
    }
    if( words(82) & 0x02 ){
        printf("\n**Security Status**\n\n");
        printf("Security Level:           %s\n", ((words(128)>>8)&0x01)?"MAXIMUM":"HIGH");
	printf("Enhanced Security Erase:  %s\n", (words(128)&0x20)?"SUPPORTED":"NOT SUPPORTED");
	printf("Security Count:           %s\n", (words(128)&0x10)?"EXPIRED":"NOT EXPIRED");
	printf("Security Frozen:          %s\n", (words(128)&0x08)?"YES":"NO");
	printf("Security Locked:          %s\n", (words(128)& 0x04)?"YES":"NO");
	printf("Security enabled:         %s\n", (words(128)& 0x02)?"YES":"NO"); 
	printf("Security supported:       %s\n", (words(128)& 0x01)?"YES":"NO"); 

    }
    if( words(84)&0x04){
        printf("\n**CFA Power Mode 1**\n\n");
        printf("Words 160:  %s\n", ((words(160)>>8)&0x80)?"SUPPORTED":"NOT SUPPORTED");
	printf("CFA Mode 1 is required for atleast one command:  %s\n", ((words(128)>>8)&0x20)?"YES":"NO");
	printf("Maxmimum Current in MA:  %d\n", words(128) & 0x0fff);
    }
   printf("\n");
}

void print_dco_identify( dco_struct* dco ){
    //print dco structure
    printf("***DCO IDENTIFY***\n");
    printf("Revision Code: %d\n", dco->revision);
    printf("Maximum Reported LBA: %llu\n", dco->max_lba[0] + ( dco->max_lba[1] <<16) 
	    						+ ( ((uint64_t)dco->max_lba[2])<<32) +( ((uint64_t)dco->max_lba[3])<<48) );

    printf("\n***Multiword DMA Support***\n\n");
    if( dco->m_dma & 0x04 ) printf("Multiword DMA 2 and below:  SUPPORTED\n");
    if( dco->m_dma & 0x02 ) printf("Multiword DMA 1 and below:  SUPPORTED\n");
    if( dco->m_dma & 0x01 ) printf("Multiword DMA 0:            SUPPORTED\n");
    
    printf("\n***Ultra DMA Support***\n\n");
    if( dco->u_dma & 0x20 ) printf("Ultra DMA 5 and below:  SUPPORTED\n");
    if( dco->u_dma & 0x10 ) printf("Ultra DMA 4 and below:  SUPPORTED\n");
    if( dco->u_dma & 0x08 ) printf("Ultra DMA 3 and below:  SUPPORTED\n");
    if( dco->u_dma & 0x04 ) printf("Ultra DMA 2 and below:  SUPPORTED\n");
    if( dco->u_dma & 0x02 ) printf("Ultra DMA 1 and below:  SUPPORTED\n");
    if( dco->u_dma & 0x01 ) printf("Ultra DMA 0:            SUPPORTED\n"); 
    
    printf("\n**Command Set/Feature Set Support**\n\n");
    printf("48-bit Addressing Feature Set:              %s\n", (dco->command_feature&0x0100)?"SUPPORTED":"NOT SUPPORTED");
    printf("Host Protected Area Feature Set:            %s\n", (dco->command_feature&0x80)?"SUPPORTED":"NOT SUPPORTED");
    printf("Automatic Acoustic Management Feature Set:  %s\n", (dco->command_feature&0x40)?"SUPPORTED":"NOT SUPPORTED");
    printf("Read/Write DMA Queued commands:             %s\n", (dco->command_feature&0x20)?"SUPPORTED":"NOT SUPPORTED");
    printf("Power Up In Standby Feature Set:            %s\n", (dco->command_feature&0x10)?"SUPPORTED":"NOT SUPPORTED");
    printf("Security Feature Set:                       %s\n", (dco->command_feature&0x08)?"SUPPORTED":"NOT SUPPORTED");
    printf("SMART Error Log:                            %s\n", (dco->command_feature&0x04)?"SUPPORTED":"NOT SUPPORTED");
    printf("SMART Self-Test:                            %s\n", (dco->command_feature&0x02)?"SUPPORTED":"NOT SUPPORTED"); 
    printf("SMART Feature Set:                          %s\n", (dco->command_feature&0x01)?"SUPPORTED":"NOT SUPPORTED");
}
